import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.rmi.RemoteException;
import javax.microedition.io.StreamConnection;
import javax.xml.rpc.Stub;
import org.jaebi.midlet.WSClient.JAEBIWSSEI;
import org.jaebi.midlet.WSClient.JAEBIWSSEI_Stub;

import org.jaebi.midlet.bt.BtServer;
/*
 * ConnectionProcessor.java
 *
 * Created on 15 giugno 2005, 23.15
 *
 * Classe incaricata di gestire la logica del message passing tra client e server
 * e, in generale, la logica con cui gestire ogni singola connessione
 */

/**
 * Classe che fa da gateway tra le midlet client e le midlet server.
 * Conosce il BtServer e la classe stub del WS. Riceve il BtServer come
 * parametro nel suo costruttore perch in caso di errore di IO durante la
 * comunicazione col client la connessione viene rimossa. Questa classe 
 * strettamente legata all'implementazione del WS, in particolare ai metodi che
 * esso espone. Attualmente sono solo 3 i possibili metodi che si possono
 * invocare:
 * <li>
 *  <ul> <b>login</b> <i>per l'autenticazione al sistema JAEBI</i></ul>
 *  <ul> <b>logout</b> </i>per il logout dal sistema jaebi</i> </ul>
 *  <ul> <b>executeRequest</b> <i>per effettuare una richiesta relativa all'
 * interazione col backEnd mappato dal sistema Jaebi</i></ul>
 * </li>
 * @author Antonio
 */
public class ConnectionProcessor implements Runnable{
    
    private StreamConnection connection = null;
    //TODO deve contenere pure la lcasse stub WS
    private int sessionId;
    
    private BtServer server;
    
    private JAEBIWSSEI service;
    
    /** */
    public ConnectionProcessor(StreamConnection connection, BtServer server, String serviceURL) {
        this.connection = connection;
        this.server = server;
        this.service = service = new JAEBIWSSEI_Stub();
        
        // Initialize the stub/service.
        ((JAEBIWSSEI_Stub)service)._setProperty(Stub.ENDPOINT_ADDRESS_PROPERTY,
                serviceURL);
         new Thread(this).start();
    }
    
    /*
     * Legge dallo stream della midlet client i dati necessari ad inoltrare una
     * richiesta al WS, ossia, ad invocare uno dei suoi metodi esposti
     * @param input lo stream proveniente dalla midlet client
     * @param endOfRequestMarker stringa indicante che sullo stream non 
     * presente alcun altra info necesaria all'invocazione di un metodo del WS
     */
    private String getRequest(DataInputStream input, String endOfRequestMarker) throws IOException{
        String temp;
        boolean isEnded = false;
        int byteReaded;
        byte buffer[] = new byte[endOfRequestMarker.length()];
        String received = "";
        
        do{
            byteReaded = input.read(buffer);
            temp = new String(buffer);
            
            received += temp.substring(0, byteReaded);
            
//            System.out.println("ConnectionProcessor:: " + received);
            
            
                /*
                 * quando  stata ricevuta la stringa che marca la fine della
                 * request (endOfRequestMarker) non devo fare pi alcuna read
                 */
            if (received.length()>= endOfRequestMarker.length() ){
                int startIndex = received.length()- endOfRequestMarker.length();
                if (received.substring(startIndex, received.length()).equalsIgnoreCase(endOfRequestMarker))
                    isEnded = true;
            }
        }
        while( !isEnded);
        
        return received;
    }
    
    
    private String getRequestContent(String request, String beginMarker, String endMarker){
        StringBuffer out = new StringBuffer(request);
        out.delete(0, beginMarker.length());
        out.delete(out.length()-endMarker.length(), out.length());
        return out.toString();
    }
    
    
    public void run(){
        boolean logout = false;
        DataInputStream streamFromClient = null;
        DataOutputStream streamToClient  = null;
        
        try{
            streamFromClient = this.connection.openDataInputStream();
            streamToClient = this.connection.openDataOutputStream();
        } catch(IOException e){
            /**
             *  avvenuto un errore durante la comunicazione: l'unica azione da
             * intraprendere  la rimozione della connessione
             */
            
            //TODO debug
            System.out.println("errore");
            
            e.printStackTrace();
            //se avviene l'eccezione ci sono errori sulla connessione:bisogna rimuoverla
            server.removeConnection(this.connection);
            return;
        }
        
        StreamConnection conn = this.connection;
        int i = 0;
        do{
            System.out.println("ConnectionProcessor::run  la " + ++i + "volta");
            try{
                
                byte buffer[] = new byte[8];
                
                String request ="";
                String response ="";
                String responseBeginMarker = "";
                String responseEndMarker = "";
                
                int result  = -1;
                int loginBeginMarkerLength = 7;
                int logoutBeginMarkerLength = 8;
                
                //flag se indica se il metodo da invocare  executeRequest
                boolean executeRequest = true;
                
                /*******************************************************************
                 * Leggo i prmi 8 caratteri che ha inviato il client per capire
                 * qual' la richiesta che si desidera inoltrare al WS, ossia, quale
                 * dei suoi metodi invocare.
                 *******************************************************************
                 */
                streamFromClient.read(buffer);
                request = new String(buffer);
                
                
            /*
             *******************************************************************
             * bisogna invocare login()?
             *******************************************************************
             */
                if (request.substring(0,loginBeginMarkerLength).equals("[login]")){
                    
                    responseBeginMarker = "[login_response]";
                    responseEndMarker = "[/login_response]";
                    
                    request += getRequest(streamFromClient, "[/login]");
                    
                    String requestContent = getRequestContent(request, "[login]", "[/login]");
                    
                    String nickName = requestContent.substring(0, requestContent.indexOf(" "));
                    String userType = "Midlet";
                    
                    response = "error";
                    
                    try{
                        result  = service.login(nickName, userType);
                        if (result != -1){
                            this.sessionId = result;
                            response="ok";
                        }
                        
                        
                        //TODO debug
                        System.out.println("ConnectionProcessor::run sessid "+ result);
                        
                        
                    } catch (RemoteException e){
                        response = "error";
                    }
                    
                    response = responseBeginMarker + response + responseEndMarker;
                    streamToClient.write(response.getBytes());
                    
                    executeRequest = false;
                    
                }
                
                
             /*
              ******************************************************************
              * bisogna invocare logout()?
              ******************************************************************
              */
                if (request.substring(0,logoutBeginMarkerLength).equals("[logout]")){
                    request += getRequest(streamFromClient, "[/logout]");
                    
                    executeRequest = false;
                    service.logout(sessionId);
                    //TODO devo mandare un response al client
                    logout = true;
                    
                }
                
                
            /*
             *******************************************************************
             * bisogna invocare xrequest()?
             *******************************************************************
             */
                if(executeRequest){
                    request += getRequest(streamFromClient, "</xrequest>");
                    //TODO debug
                    System.out.println("ConnectionProcessor:: run xrequest:" + request);
                    response = service.executeRequest(this.sessionId, request);
                    
                    //TODO debug
                    System.out.println("ConnectionProcessor:: run xresponse:" + response);
                    streamToClient.write(response.getBytes());
                    
                    //TODO devo controllare se oltre i marker non c' nient'altro?
                    
                }
                
            }
            
            catch(IOException e){
                /**
                 *  avvenuto un errore durante la comunicazione: l'unica azione da
                 * intraprendere  la rimozione della connessione
                 */
                
                //TODO debug
                System.out.println("errore");
                
                e.printStackTrace();
                //se avviene l'eccezione ci sono errori sulla connessione:bisogna rimuoverla
                server.removeConnection(this.connection);
            }
        }
        while (!logout);
        
    }
    
    
}
