/*
 * ConnectionObserver.java
 *
 * Created on 14 giugno 2005, 22.48
 *
 * Rappresenta un Observer per la classe ConnectionSet, il pool di connessioni 
 * attive sul Server. L'aggiunta o la rimozione di una connessione rappresenta 
 * un "evento", che viene notificato a questa classe attraverso una chiamata al 
 * suo metodo update. All'interno di update si identifica l'evento che ha 
 * scatenato la notifica e si richiama l'opportuno handler, ossia, un metodo 
 * dell'interfaccia ConnectionEventHandler
 */

package org.jaebi.midlet.bt;

import org.jaebi.midlet.util.Observable;
import org.jaebi.midlet.util.Observer;

/**
 *
 * @author Antonio
 */
public class ConnectionObserver implements Observer{
    ConnectionEventHandler handler;
    
    /** Creates a new instance of ConnectionObserver */
    public ConnectionObserver(ConnectionEventHandler handler) {
        this.handler = handler;
    }
    
    public void update(Observable o, Object arg){
        ConnectionSet connectionSet = (ConnectionSet) o;
        ConnectionEvent event = (ConnectionEvent) arg;
        
        switch(event.getEvent()){
            case ConnectionEvent.CONNECTION_ADDED:
                handler.connectionAdded(event.getConnection());
            break;
            
            case ConnectionEvent.CONNECTION_REMOVED:
                handler.connectionRemoved(event.getConnection());
            break;
        }
        
    }
    
}
