/*
 * Discoverer.java
 *
 * Created on 30 maggio 2005, 15.58
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 *TODO deve essere un singleton..o no? cio se lo faccio come singleton, poi posso cmq fare in modo che due applicazioni diverse in esecuzione nello stesso momento usino 2 discoverer differenti?
 */

package org.jaebi.midlet.bt;
import java.util.Enumeration;
import java.util.Hashtable;
import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DeviceClass;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.DiscoveryListener;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;
import javax.bluetooth.UUID;
import org.jaebi.midlet.bt.btException.BtInitException;
import org.jaebi.midlet.bt.btException.DiscoveryException.DiscoveryInterruptionException;
import org.jaebi.midlet.bt.btException.DiscoveryException.InquiryErrorException;
import org.jaebi.midlet.bt.btException.DiscoveryException.StartInquiryException;
import org.jaebi.midlet.bt.btException.DiscoveryException.UnexpectedDiscoveryCodeException;
import org.jaebi.midlet.bt.btException.DiscoveryException.UnexpectedInterruptionException;
import org.jaebi.midlet.bt.btException.DiscoveryException.UserInterruptionException;
import org.jaebi.midlet.util.Observable;
import org.jaebi.midlet.util.Observer;
import org.jaebi.midlet.bt.discoveryHandling.DiscoveryReport;
import org.jaebi.midlet.bt.discoveryHandling.ServiceSearchError;
import org.jaebi.midlet.bt.discoveryHandling.DeviceSearchError;
import org.jaebi.midlet.bt.discoveryHandling.DiscoveryError;

/**
 * Ricerca delle device bluetooth a cui collegarsi e/o servizi di cui usufruire
 * Implementa l'Observable del pattern Observer - Observers. Il Client
 * (o i Client) che utilizzeranno questa classe, invece, rappresenteranno gli
 * Observer.
 * La notifica ai propri Observer viene inviata ogni qual volta avviene un
 * cambiamento della variabile state; in particolare si distinguono due
 * cambiamenti di stato, a seconda se il cambiamento  dovuto a causa di un
 * errore o meno:
 *<ul>
 *  <li>
 *      Se lo stato passa a <code>READY</code>, <code>DEVICE_SEARCH</code> o
 *      <code>SERVICE_SEARCH</code> si effettua una chiamata a notifyObservers()
 *      e gli observers recuperano lo stato attuale attraverso il metodo getState()
 * </li>
 * <li>
 *      Se lo stato passa a <code>DEVICE_SEARCH_ERROR</code>
 *      avviene una chiamata a
 *      notifyObservers(discoveryError) il cui argomento rappresenta un'istanza
 *      della classe DiscoveryError
 * </li>
 * <li>
 *      Se lo stato passa a <code>NEW_SERVICE_SEARCH_ERROR</code> avviene una chiamata a
 *      notifyObservers(serviceError) il cui argomento rappresenta un'istanza
 *      della classe ServiceError
 * </li>
 *      Se lo stato passa a <code>SERVICE_SEARCH_COMPLETE</code> avviene una
 *      chiamata a notifyObservers(rd), con rd istanza della device per la quale
 *       terminata la ricerca di servizi
 * <li>
 *
 * </li>
 *</ul>
 *
 *@see org.jaebi.client.midlet.bt.DiscoveryError
 *
 * TODO quando si fa una chiamata a qualche cancel lo stato dovrebbe passare a
 * READY
 *
 * @author Antonio
 */
public class Discoverer extends Observable implements DiscoveryListener, DiscovererState, Runnable {
    
    // riferimento alla local Device relativa alla periferica bluetooth su cui  in esecuzione il discoverer
    //TODO commentare meglio (cos' una LocalDevice?vedi javadoc)
    //TODO forse non  necessario che sia una variabile di istanza()probabilmente verr usata solo nel costruttore per prendere il discovery
    LocalDevice localDevice;
    
    //riferimento al discovery agent relativo alla device bluetooth su cui  in esecuizione il discoverer
    //TODO commentare meglio (cos' una discoveryAgent? vedi javadoc)
    private DiscoveryAgent discoveryAgent;
    
    /** tiene traccia dello stato corrente del discoverer. */
    private int state = READY;
    
    /** tiene traccia delle device trovate durante l'inquiry e, per ogni device,
     * dei servizi che offre
     */
    private DiscoveryReport discoveryReport;
    
    /** Ogni qual volta si inizia una ricerca dei servizi disponibili su una
     * determinata device, il sistema associa un id a tale richiesta.
     * Questa var associa un id di ricerca servizi con la device su cui tale
     * ricerca  stata sottomessa.
     * siccome si effettua una sola ricerca di servizi per ogni device
     * trovata durante l'inquiry, questa hashTable contiene tanti id quante sono
     * le device trovate tramite searchDevice() (ossia esiste un'associazione 1:1
     * tra device torvate e id di ricerca)
     * Ogni elemento di questa hashtable  quindi un riferimento ad una device
     * bluetooth, memorizzato in chiave dell'id di ricerca
     * servizi a cui la device  associata
     */
    private Hashtable serviceSearches;
    
    /** tiene traccia del codice ritornato dalla ricerca delle device */
    private int discType;
    
    private boolean closed;
    
    /** insieme degli UUID che un servizio deve avere per essere ritrovato da
     * questo discoverer
     */
    private UUID UUIDset[];
    
    /** insieme di attributi che un servizio deve avere per essere ritrovato da
     *  questo discoverer
     */
    private int attrset[];
    /** Creates a new instance of Discoverer */
    
    /* TODO forse  necessario implementarlo come thread perch la wait in searchDevices() potrebbe
     * fare andare tutto in deadlock (se il thread principale si sospende pu essere che la chiamata del sistema a inquiryCompleted()
     * non ha effetto e quindi non avviene mai la chiamata a notify
     */
    public Discoverer(DiscovererObserver discovererObserver, UUID UUIDset[], int attrset[] ) throws BtInitException{
        closed = false;
        this.discoveryReport = new DiscoveryReport();
        
        try{
            // crea/ottiene un localDevice e un discovery agent
            localDevice = LocalDevice.getLocalDevice();
            discoveryAgent = localDevice.getDiscoveryAgent();
        }catch (BluetoothStateException e1) {
            throw new BtInitException("Impossibile inizializzare la periferica bluetooth");
        }
        this.addObserver((Observer)discovererObserver);
        this.UUIDset = UUIDset;
        this.attrset = attrset;
    }
    
    public Discoverer(DiscovererObserver discovererObservers[], UUID UUIDset[], int attrset[])throws BtInitException{
        closed = false;
        this.discoveryReport = new DiscoveryReport();
        
        try{
            // crea/ottiene un localDevice e un discovery agent
            localDevice = LocalDevice.getLocalDevice();
            discoveryAgent = localDevice.getDiscoveryAgent();
        }catch (BluetoothStateException e1) {
            throw new BtInitException("Impossibile inizializzare la periferica bluetooth");
        }
        
        for (int i =0; i<discovererObservers.length; i++){
            this.addObserver((Observer)discovererObservers[i]);
        }
        this.UUIDset = UUIDset;
        this.attrset = attrset;
    }
    
    
    public boolean isClosed(){
        return this.closed;
    }
    
    
    public int getState(){
        return this.state;
    }
    
    
    private void setState(int state){
        this.state = state;
        setChanged();
    }
    
    
    /**
     * Ricerca nuove device bluetooth a cui collegarsi (effettua, in pratica,
     * lo start dell'inquiry). Il metodo  dichiarato synchronized per evitare
     * che pi di un flusso di controllo effettui concorrentemente la ricerca
     * delle device
     */
    private synchronized void searchDevices() throws StartInquiryException, DiscoveryInterruptionException, InquiryErrorException, UnexpectedDiscoveryCodeException{
        
        try {
            discoveryAgent.startInquiry(DiscoveryAgent.GIAC, this);
        } catch (BluetoothStateException e) {
            setState(DISCOVERY_ERROR);
            throw new StartInquiryException("Impossibile iniziare l'inquiry; impossibile iniziare la ricerca di device");
        }
        
        try {
            // sospendo il thread finch non viene terminata la ricerca delle device (cio finch il sistema non effettua la chiamata a inquiryCompleted()
            wait();
        } catch (InterruptedException e) {
            setState(DISCOVERY_ERROR);
            throw new UnexpectedInterruptionException("Ricerca device terminata a causa di un'interruzione inaspettata");
        }
        
        // il thread  stato risvegliato per una chiamata a destroy()?
        if (isClosed()) {
            setState(DISCOVERY_ERROR);
            throw new DiscoveryInterruptionException("Discovery interrotto");
        }
        
        /* no? Allora a questo punto il sistema ha terminato l'inquiry, avr invocato
         * inquiryCompleted() che a sua volta avr ricopiato il codice di
         * ritorno nella var di istanza discType
         */
        switch (discType) {
            case INQUIRY_ERROR:
                setState(DISCOVERY_ERROR);
                throw new InquiryErrorException("errore durante la ricerca di device");
            case INQUIRY_TERMINATED:
                
                /*
                 * Se si  arrivati qui, significa che  stata fatta una chiamata
                 * a cancelSearch()
                 */
                setState(DISCOVERY_ERROR);
                throw new UserInterruptionException("Ricerca delle device interrotta");
                
            case INQUIRY_COMPLETED:
                
                break;
            default:
                setState(DISCOVERY_ERROR);
                //discType sconosciuto
                destroy();
                throw new UnexpectedDiscoveryCodeException("Errore di sistema: discovery code sconosciuto");
        }
    }
    
    
    /**
     * E' da notare che, differentemente da quanto accade all'interno di
     * searchDevice(), lo switch sul respCode ritornato dalla ricerca dei
     * servizi viene fatto direttamente all'interno dell'handler serviceSearchCompleted().
     * Questo perch ho interesse che gli observer siano informati in tempo reale
     * dell'esito della ricerca su una determinata device. Siccome, nel nostro caso,
     * il thread che invoca searchServices si mette in attesa finch tutte le
     * ricerche di servizi non vengano completate, se si effettuava il controllo
     * qui si potevano informare gli observer solo alla fine di tutte le ricerche,
     * non ogni qual volta ne terminasse una
     */
    private synchronized void searchServices() throws DiscoveryInterruptionException{
        serviceSearches = new Hashtable();
        Enumeration devices = discoveryReport.getDevices();
        DiscoveryError discoveryError = null;
        
        //TODO debug
        System.out.println("sono il client e sto cercando questi UUID");
        for (int i=0; i<UUIDset.length; i++)
            System.out.println(UUIDset[i]);
        System.out.println("sono il client e sto cercando questi attr");
        for (int i=0; i<attrset.length; i++)
            System.out.println(attrset[i]);
        //fine debug
        
        /**
         * Ricerco i servizi disponibili sulle device trovate attraverso
         * l'inquiry. Invio una richiesta di ricerca per ogni device trovata.
         * Tutte le ricerche avvengono concorrenzialmente
         */
        while (devices.hasMoreElements()) {
            RemoteDevice rd = (RemoteDevice) devices.nextElement();
            //TODO debug
            System.out.println("sono il client ed ho indirizzo " + localDevice.getBluetoothAddress() );
            System.out.println("sono il client ed ho trovato un server con indirizzo " + rd.getBluetoothAddress() );
            
            try {
                int id = discoveryAgent.searchServices(null, UUIDset,rd, this);
                serviceSearches.put(new Integer(id), rd );
            } catch (BluetoothStateException e) {
                /**
                 * Non faccio lanciare alcuna eccezione perch se su una device
                 * non  possibile ricercare i servizi disponibili, voglio che
                 * la ricerca continui per le altre device
                 */
                
                String errorMsg = "Impossibile ricercare i servizi offerti dalla device ";
                
                discoveryError = new ServiceSearchError(ServiceSearchError.CANT_START_SEARCH, errorMsg, rd);
                setState(NEW_SERVICE_SEARCH_ERROR);
                notifyObservers(discoveryError);
                setState(SERVICE_SEARCH);
                notifyObservers();
            }
            
            try {
                /*
                 * sospendo il thread finch tutte le ricerche di servizi non
                 * siano terminate
                 */
                wait();
            } catch (InterruptedException e) {
                setState(DISCOVERY_ERROR);
                throw new UnexpectedInterruptionException("Ricerca servizi terminata a causa di un'interruzione inaspettata");
            }
            
            // il thread  stato risvegliato per una chiamata a destroy()?
            if (isClosed()) {
                setState(DISCOVERY_ERROR);
                throw new DiscoveryInterruptionException("Discovery interrotto");
            }
            
             /* no? Allora a questo punto il sistema ha terminato la ricerca,
              * avr invocato ServiceSearchCompleted() -in particolare avr
              * fatto tante chiamate a questo metodo quante sono le ricerche di
              * servizi richieste- che a sua volta avr opportunamente "riempito"
              * respCode
              */
        }
    }
    
    
    public void run(){
        discoveryReport.clear();
        DiscoveryError discoveryError = null;
        
        /*_____________________________________________________________________
         * comincio la scansione delle device
         * in caso di errore, istanzio opportunamente deviceError e chiamo
         * setChanged per la notifica agli observer
         *_____________________________________________________________________
         */
        setState(DEVICE_SEARCH);
        notifyObservers();
        //TODO nei catch di qui sotto deve andare un return (che continuo a fare se la ricerca delle device ha dato errore?
        //TODO nei casi di errori gravi, svuotare discoveryreport
        try{
            searchDevices();
        }
        
        catch(StartInquiryException e){
            discoveryError = new DeviceSearchError(DeviceSearchError.CANT_START_INQUIRY, e.getMessage());
        }
        
        catch (DiscoveryInterruptionException e1){
            
            /***********************************************************
             * l'inquiry  stato interrotto; sono 3 le possibili cause:
             ***********************************************************/
            
            String errorMsg = e1.getMessage();
            Exception ex;
            
            try{
                
                /*************************************************
                 * 1)  stata fatta una chiamata a cancelSearch()
                 *************************************************/
                
                ex =(UserInterruptionException)e1;
                discoveryError = new DeviceSearchError(DeviceSearchError.INQUIRY_TERMINATED, errorMsg);
            } catch (ClassCastException ex1){
                
                try{
                    
                    /**************************************************
                     * 2) la wait ha lanciato un InterruptedException
                     **************************************************/
                    
                    ex = (UnexpectedInterruptionException)e1;
                    discoveryError = new DiscoveryError(DeviceSearchError.UNEXPECTED_DISCOVERY_INTERRUPTION, errorMsg);
                } catch (ClassCastException ex2){
                    
                    /******************************************
                     * 3)  stata fatta una chiamata a destroy()
                     ******************************************/
                    
                    discoveryError = new DiscoveryError(DeviceSearchError.DISCOVERY_INTERRUPTION, errorMsg);
                }
                
            }
            
        }
        
        catch (InquiryErrorException e2){
            discoveryError = new DeviceSearchError(DeviceSearchError.INQUIRY_ERROR, e2.getMessage());
        }
        
        catch(UnexpectedDiscoveryCodeException e3){
            discoveryError = new DiscoveryError(DeviceSearchError.UNKNOWN_DISCOVERY_CODE, e3.getMessage());
        }
        
        notifyObservers(discoveryError);
        
        /**
         * se si  verificato errore o se non  stata trovata alcuna periferica,  inutile continuare
         */
        if( (discoveryError != null) || (discoveryReport.deviceCount()==0) ){
            setState(READY);
            notifyObservers();
            return;
        }
        
        /*______________________________________________________________________
         * Comincio la scansione dei servizi
         *______________________________________________________________________
         */
        setState(SERVICE_SEARCH);
        
        /*
         * notifico agli observer che lo stato del discoverer  cambiato ed
         * invio loro la lista delle devices trovate
         */
        notifyObservers();
        
        try{
            searchServices();
        }
        
        catch (DiscoveryInterruptionException e) {
            Exception ex;
            String errorMsg = e.getMessage();
            try{
                ex = (UnexpectedInterruptionException)e;
                discoveryError = new DiscoveryError(DiscoveryError.UNEXPECTED_DISCOVERY_INTERRUPTION, errorMsg);
            } catch (ClassCastException e1){
                discoveryError = new DiscoveryError(DiscoveryError.DISCOVERY_INTERRUPTION, errorMsg);
            }
        }
        
        notifyObservers(discoveryError);
        
        setState(READY);
        notifyObservers();
        
    }
    
    
    /**
     * permette di effettuare una ricerca di device e servizi
     **/
    public void scanDevicesServices(){
        closed = false;
        new Thread(this).start();
    }
    
    
    /**
     * Metodo richiamato automaticamente dal sistema quando viene trovata una
     * periferica bluetooth in attesa di connessioni
     */
    public void deviceDiscovered(RemoteDevice btDevice,
            DeviceClass cod){
        
        discoveryReport.addDevice(btDevice);
        
        //TODO debug
        System.out.println("device trovata: " + btDevice.getBluetoothAddress());
    }
    
    
    /**
     * Metodo richiamato dal sistema quando viene completato un inquiry
     */
    public void inquiryCompleted(int discType){
        
        /*
         * Ricopio il discType ritornato dal sistema nell'opportuna var di
         * istanza
         */
        this.discType = discType;
        
        /* risveglio il thread (lanciato da scanDevicesServices()) che era in
         * attesa del completamento dell'inquiry
         */
        //TODO capire perch  una reg critica
        synchronized (this) {
            notify();
        }
    }
    
    
    /**
     * Metodo richiamato automaticamente dal sistema quando viene trovato un
     * servizio durante la ricerca dei servizi
     */
    public void servicesDiscovered(int transID,ServiceRecord[] servRecord){
        //TODO debug
        System.out.println("servizio scoperto");
        
        for (int i = 0; i < servRecord.length; i++) {
            discoveryReport.addService(servRecord[i]);
            
        }
    }
    
    
    //TODO deve essere synchronized?
    /** Cancel's the devices/services search. */
    public void cancelSearch() {
        synchronized (this) {
            if (state == DEVICE_SEARCH) {
                discoveryAgent.cancelInquiry(this);
            } else if (state == SERVICE_SEARCH) {
                Enumeration en = serviceSearches.keys();
                while (en.hasMoreElements()){
                    int id = ((Integer)en.nextElement()).intValue();
                    discoveryAgent.cancelServiceSearch(id);
                }
            }
        }
    }
    
    
    //TODO deve essere synchronized?
    public void cancelServiceSearch(int id){
        discoveryAgent.cancelServiceSearch(id);
    }
    
    
    //TODO deve essere synchronized?
    public void cancelAllPendigServiceSearch(){
        Enumeration en = serviceSearches.keys();
        
        while (en.hasMoreElements()){
            int id = ((Integer)en.nextElement()).intValue();
            discoveryAgent.cancelServiceSearch(id);
        }
    }
    
    /**
     * Metodo richiamato automaticamente dal sistema quando viene terminata la
     * ricerca dei servizi
     * [vedi commento a searchServices]
     */
    public void serviceSearchCompleted(int transID, int respCode){
        boolean error = false;
        String errorMsg;
        
        /*
         * Recupero la device su cui  stata sottomessa la ricerca con
         * id = transid
         */
        Integer id = new Integer(transID);
        RemoteDevice rd = (RemoteDevice) serviceSearches.get(id);
        
        //TODO debug
        if (rd != null){
            System.out.println(" finita la ricerca di servizi per la device con indirizzo " + rd.getBluetoothAddress());
            System.out.println("ci sono " + discoveryReport.getServices(rd).size()+" servizi associati a questa device");
        } else
            System.out.println("trans id inatteso");
        
        DiscoveryError discoveryError = null;
        
        
        if (!serviceSearches.containsKey(id)) {
            /*******************************************************************
             * error - l'id della ricerca appena terminata non era tra gli id
             * delle nostre richieste
             ******************************************************************/
            
            error = true;
            errorMsg = "Id ricerca non previsto";
            setState(NEW_SERVICE_SEARCH_ERROR);
            discoveryError = new ServiceSearchError(ServiceSearchError.UNEXPECTED_TRANS_ID, errorMsg, rd);
            
            // TODO devo far fare qualcos'altro?
        }
        
        //aggiorno la lista delle ricerche in pending
        this.serviceSearches.remove(id);
        
        
        /*
         * se l'id della ricerca terminata non  tra quelle pending,  inutile
         * che controllo il suo respCode
         */
        if (!error){
            switch (respCode){
                
                case SERVICE_SEARCH_ERROR:
                    setState(NEW_SERVICE_SEARCH_ERROR);
                    errorMsg = "Errore durante la ricerca di servizi";
                    discoveryError = new ServiceSearchError(ServiceSearchError.SEARCH_ERROR, errorMsg, rd);
                    break;
                    
                case SERVICE_SEARCH_TERMINATED:
                /*
                 * Se si  arrivati qui, significa che  stata fatta una chiamata
                 * a cancelServiceSearch() o cancelAllPendingSrviceSearch
                 */
                    setState(NEW_SERVICE_SEARCH_ERROR);
                    errorMsg = "Ricerca dei servizi interrotta";
                    discoveryError = new ServiceSearchError(ServiceSearchError.SEARCH_TERMINATED, errorMsg, rd);
                    break;
                    
                case SERVICE_SEARCH_COMPLETED:
                    setState(this.NEW_SERVICE_SEARCH_COMPLETE);
                    
                    /* chiamo un notifyObservers perch voglio che sia inviato
                     * loro la device su cui  terminata la ricerca cos poi attravesro
                     *observable recupero i servizxi
                     */
                    //TODO perch mi da class cast exception???
                    notifyObservers(rd);
                    break;
                    
                case SERVICE_SEARCH_NO_RECORDS:
                    setState(NEW_SERVICE_SEARCH_ERROR);
                    errorMsg = "Per questa device non  stato trovato alcun record di servizio compatibile con quelli richiesti";
                    discoveryError = new ServiceSearchError(ServiceSearchError.SEARCH_TERMINATED, errorMsg, rd);
                    break;
                    
                case SERVICE_SEARCH_DEVICE_NOT_REACHABLE:
                    setState(NEW_SERVICE_SEARCH_ERROR);
                    errorMsg = "Per questa device non  possibile continuare la ricerca di servizi perch la device risulta non raggiungibile";
                    discoveryError = new ServiceSearchError(ServiceSearchError.SEARCH_TERMINATED, errorMsg, rd);
                    break;
                    
                default:
                    setState(NEW_SERVICE_SEARCH_ERROR);
                    errorMsg = "Response code sconosciuto";
                    discoveryError = new ServiceSearchError(ServiceSearchError.UNKNOWN_DISCOVERY_CODE, errorMsg, rd);
                    break;
            }
        }
        
        
        /**
         * notifico gli observer circa il nuovo stato del discoverer; Se la
         * ricerca ha prodotto qualche errore, invio loro un'istanza di
         * ServiceSearchError, altrimenti l'argomento di notifyObservers
         * sar null
         */
        
        notifyObservers(discoveryError);
        
        /*
         ***************************************************************
         * controllo che la ricerca terminata sia l'ultima sottomessa
         ****************************************************************
         */
        
        /*
         * Se c' ancora almeno una ricerca che deve terminare, reimposto lo
         * stato a SERVICE_SEARCH ed effettuo un return. Le operazioni
         * successive a questo if verranno effettuate solo
         * nel momento in cui termina l'ultima ricerca di servizi sottomessa
         */
        if (serviceSearches.size() > 0) {
            setState(this.SERVICE_SEARCH);
            notifyObservers();
            return;
        }
        
        /*
         * ok, tutte le ricerche di servizi richieste sono state completate
         * risveglio il thread sospeso che, dopo aver effettuato i controlli
         * sull'eventuale occorrenza di errori, imposter
         * lo stato dell discoverer a READY
         */
        synchronized (this) {
            //risveglio il thread sospeso
            notify();
        }
    }
    
    
    public DiscoveryReport getDiscoveryReport(){
        return discoveryReport;
    }
    
    
    /**
     * Annulla qualsiasi ricerca di servizi o device richiesta al sistema
     */
    public void destroy(){
        /* risveglio l'eventuale thread in attesa che la scansione delle device
         * o dei servizi termini
         */
        notify();
        cancelSearch();
        closed = true;
        //TODO implementami!!
    }
    
}
