/*
 * Listener.java
 *
 * Created on 26 maggio 2005, 11.38
 *
 * Gestore di connessioni ad una bluetooth master device.
 * Implementa un thread che si mette in attesa di nuove connessioni bluetooth.
 * L'utilizzo di un thread  giustificato dal fatto che si vuole evitare che l'intera applicazione
 * che usa Listener si blocchi sul metodo acceptAndOpen() nel caso in cui non vi siano connessioni
 * TODO deve essere un sinlgeton!!
 *
 */

package org.jaebi.midlet.bt;

import java.io.IOException;
import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DataElement;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;
import javax.microedition.io.StreamConnectionNotifier;
import org.jaebi.midlet.bt.btException.BtInitException;


/**
 * Classe incaricata di rilevare le connessioni in entrata ad 
 * un server bluetooth
 * @author Antonio
 */
public class Listener{
    
    // url del server bluetooth
    private StringBuffer url;
    private StreamConnectionNotifier notifier;
    
    //incapsula le informazioni riguardo il servizio offerto da questo server BT
    private ServiceRecord record;
    
    // riferimento alla local Device relativa alla periferica bluetooth su cui  in esecuzione il listner
    //TODO commentare meglio (cos' una LocalDevice?vedi javadoc)
    private LocalDevice localDevice;
    
    //flag che indica se il server  disposto o meno ad accettare connessioni
    private boolean closed;
    
    /**
     * Istanzia un listner settando la device BT su cui  in esecuzione come 
     * discoverable  con inquiry access code General/Unlimited Inquiry Access Code.
     * @see javax.bluetooth.DiscoveryAgent
     * @see javax.bluetooth.LocalDevice
     * TODO un possibile sviluppo futuro potrebbe rendere il costruttore parametrico anche in base all'
     * inquiry mode
     * 
     * @param url rappresenta l'url del server. Tale url deve essere conforme al seguente formato
     *{scheme}:[{target}][{params}]
     * dove:
     * <ul>
     *  <li> scheme rappresenta il protocollo da utilizzare</li>
     *  <li> target rappresenta l'indirizzo dell'host
     *  <li> params eventuali parametri da passare all'url
     * </ul> 
     * A esempio, se vogliamo creare un listner che accetta connessioni usando il protocollo RFCOMM, l'url
     * sar del tipo
     * <code>btspp://localhost:UUID;name=MY_SERVER;authorize=true</code>
     * se vogliamo usare il protocollo L2CAP, l'url sar del tipo:
     * <code>btl2cap://localhost:UUID;name=MY_SERVER;authorize=true</code>
     * se vogliamo, invece, usare il protocollo OBEX l'url sar del tipo
     * <code>btgoep://localhost:UUID;name=MY_SERVER;authorize=true</code>
     * @see  javax.microedition.io.Connector
     */
    //TODO overloadare questo costruttore per fargli prendere un DataElement (nel caso in cui qualcuno volesse settare degli attributi)
    public Listener(StringBuffer url) throws BtInitException {
        this.url = url;
        
        try{
            
            // memorizzo il riferimento alla local Device
            localDevice = LocalDevice.getLocalDevice();
            
            // imposto il discoverable mode per la local device
            if (!localDevice.setDiscoverable(DiscoveryAgent.GIAC)) {
                throw new BtInitException("Can't set discoverable mode...");
            }
            
            //creo il service record
            notifier = (StreamConnectionNotifier) Connector.open(url.toString());
            
            //TODO debug 
            System.out.println("url che passo al notifier:" + url.toString());
        } 
        
        catch (BluetoothStateException e1) {
            throw new BtInitException("Impossibile inizializzare la periferica bluetooth");
            
        }
        
        catch(IOException e){
            throw new BtInitException("Impossibile inizializzare la periferica bluetooth");
        }
        
        //TODO debug
        System.out.println("sono il server ed ho indirizzo " + localDevice.getBluetoothAddress() );
        
        //recupero il Service record creato sopra
        record = localDevice.getRecord(notifier);
        
        //TODO bebug
        System.out.println("sono il server e questo  l'indirizzo per connettersi"+ record.getConnectionURL(ServiceRecord.NOAUTHENTICATE_NOENCRYPT, false)); 
        
            
        //TODO debug
        int attr[];
        attr = record.getAttributeIDs();
        for (int i =0; i<attr.length; i++){
            System.out.println(attr[i]);
        }
        System.out.println((String)(record.getAttributeValue(256).getValue()));
                
        closed = false;
    }
    
    /**
     * Si mette in attesa di nuove connessioni. Non appena una nuova device si collega al server, ritorna
     * l'handle alla connessione creata
     */
    public StreamConnection listen(){
        StreamConnection conn = null;
        try {
            conn = notifier.acceptAndOpen();
        } catch (IOException e) {
            // wrong client or interrupted - continue anyway
            ;
        }
        
        String remoteDeviceName ="";
        try{
            remoteDeviceName = RemoteDevice.getRemoteDevice(conn).getFriendlyName(true);
        }catch (IOException e){
           ;
        }
        //TODO debug
        System.out.println("Listner::Si  collegata la device con nome " + remoteDeviceName);
        return conn;
        
    }
    
    public boolean isClosed(){
        return closed;
    }
    
}
