/**
 * BtClient.java
 *
 * Created on 9 giugno 2005, 16.28
 */

package org.jaebi.midlet.bt;


import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import javax.bluetooth.ServiceRecord;
import javax.bluetooth.UUID;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;
import org.jaebi.midlet.bt.btException.BtIOException;
import org.jaebi.midlet.bt.btException.BtInitException;
import org.jaebi.midlet.bt.btException.ConnectionException;
import org.jaebi.midlet.bt.discoveryHandling.DiscoveryEventHandler;

/**
** Questa classe fornisce gli strumenti necessari per implementare e quindi
     * gestire un client bluetooth. Istanzia un discoverer per la ricerca di device
     * bluetooth che offrano determinati servizi, in particolare quelli con gli UUID
     * e gli id di attributo passati al costruttore.
     * Il discoverer viene monitorato da un discovererObserver che questa classe si
     * incarica di istanziare.
     * @see Discoverer
     *
 * @author Antonio
 */
public class BtClient {
    
    private Discoverer discoverer;
    private DiscovererObserver discovererObserver;
    private DiscoveryEventHandler eventHandler;
  
    public BtClient(UUID UUIDset[], int attrset[], DiscoveryEventHandler eventHandler) throws BtInitException {
        this.eventHandler = eventHandler;
        this.discovererObserver = new DiscovererObserver(eventHandler);
        this.discoverer = new Discoverer(discovererObserver, UUIDset, attrset);
    }
    
    
    public void startDiscovery(){
        discoverer.scanDevicesServices();
    }
    
    
    public void cancelDiscovery(){
        discoverer.destroy();
    }
    
    /**
     * crea una nuova connessione per questo client al servizio identificato dal
     * parametro record
     */
    public StreamConnection getConnection(ServiceRecord record, int requiredSecurity, boolean mustBeMaster) throws ConnectionException{
        String url = record.getConnectionURL(requiredSecurity,  mustBeMaster);
        try{
            return (StreamConnection)Connector.open(url);
        } catch (IOException e){
            throw new ConnectionException("Impossibile connettersi al servizio desiderato");
        }
    }
    
    
}
