/*
 * ConnectionSet.java
 *
 * Created on 29 maggio 2005, 12.50
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package org.jaebi.midlet.bt;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Vector;
import javax.bluetooth.RemoteDevice;
import javax.microedition.io.StreamConnection;
import org.jaebi.midlet.bt.btException.BtIOException;
import org.jaebi.midlet.util.Observable;
import org.jaebi.midlet.util.Observer;

/**
 * Questa classe rappresenta un Container delle connessioni effettuate ad una Server Midlet.
 * E' un observable che notifica ai suoi observer l'aggiunta o la rimozione di una connessione.
 *
 * L'istanza di questa classe potrebbe rappresentare una regione critica per l'accesso
 * concorrente che vi fanno il thread lanciato da BtServer e il thread lanciato da ClientProcessor.
 * Dichiarando alcuni metodi come synchronized si garantisce la mutua esclusione ad essi e si evitano, quindi, possibili
 * race conditions
 *(un possibile scenario in cui si potrebbe avere una race condition  il seguente:
 * ConnectionProcessor thread rimuove una connessione per un errore di IO e
 * concorrenzialmente BtServer falseil get di quella stessa connessione
 * @author Antonio
 */
public class ConnectionSet  extends Observable{
    
    private Vector connections;
    private int event;
    
    /** Creates a new instance of ConnectionSet */
    public ConnectionSet(ConnectionObserver observer) {
        connections = new Vector();
        this.addObserver((Observer)observer);
    }
    
    
    public synchronized void addConnection(StreamConnection conn){
        ConnectionEvent event = new ConnectionEvent(ConnectionEvent.CONNECTION_ADDED, conn);
        
        connections.addElement(conn);
        this.setChanged();
        notifyObservers(event);
    }
    
    
    public synchronized void removeConnection(StreamConnection conn){
        ConnectionEvent event = new ConnectionEvent(ConnectionEvent.CONNECTION_REMOVED, conn);
        
        try{
            conn.close();
        } catch (IOException e){
            System.err.println("Errore durante la chiusura della connessione");
        }
        
        connections.removeElement(conn);
        this.setChanged();
        notifyObservers(event);
        
    }
    
    public synchronized StreamConnection getConnection(int index){
        return (StreamConnection)connections.elementAt(index);
    }
    
    /**
     * Ritrova le connessioni relative ad una certa remote device
     */
    public synchronized StreamConnection[] getConnection(RemoteDevice rd){
        Vector outTemp = new Vector();
        StreamConnection out[];
        Enumeration en = connections.elements();
        
        while (en.hasMoreElements()){
            StreamConnection conn = (StreamConnection)en.nextElement();
            RemoteDevice temp = null;
            
            try{
                temp = RemoteDevice.getRemoteDevice(conn);
                if (temp.equals(conn))
                    outTemp.addElement(conn);
            } 
            catch(IOException e){
                ;
                /* non faccio nulla perch non posso compromettere la ricerca se 
                 * per una certa connessione non riesco a recuperare la 
                 * remoteDevice
                 */
            }

        }
        
        out = new StreamConnection[outTemp.size()];
        
        for (int i=0; i<outTemp.size(); i++){
            out[i] = (StreamConnection)outTemp.elementAt(i);
        }
        
        outTemp.removeAllElements();
        return out;
    }
    
    
    /**
     *ritorna il numero di connessioni non pendenti
     */
    public synchronized int connectionsCount(){
        return connections.size();
    }
    
    /**
     *Chiude tutte le connessioni del set e svuota il Vector connections
     */
    public synchronized void closeAll() throws BtIOException{
        Enumeration en;
        en = connections.elements();
        while(en.hasMoreElements()){
            try{
                ((StreamConnection)en.nextElement()).close();
            } catch(IOException e){
                throw new BtIOException();
            }
        }
        connections.removeAllElements();
    }
    
    
}
