package org.jaebi.midlet.bt.requestProcessing;
import com.sun.midp.io.BufferedConnectionAdapter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import javax.microedition.io.StreamConnection;
import org.jaebi.midlet.bt.btException.BtIOException;
import org.kxml.parser.XmlParser;
import org.jaebi.midlet.bt.responseHandling.ServiceResponseHandler;
/*
 * RequestProcessor.java
 *
 * Created on 18 giugno 2005, 15.58
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

/**
 * questa classe incapsula tutte le operazioni per necessarie affinch una
 * midlet client invii una richiesta alla midlet server e da quest'ultima ne ri-
 * ceva il response. Per richiesta si intende un operazione da eseguire sul back
 * end per conto dell'utente che usa la midlet cliet e che si autentica al
 * sistema JAEBI er le operazioni che esegue questa classe sono sostanzialmente
 * queste:
 * <li>
 *  <ul>incapsulamento della richiesta in un opportuno envelop XML (che verr
 *      inviato dalla midlet server al WS di JAEBI)  </ul>
 *  <ul> invio della richiesta nel formato xml alla midlet Server </ul>
 *  <ul> attesa del response relativo alla richiesta da parte della midlet
 *       server (che in sostanza forwarda quello che riceve dal WS) ed esecuzione
 *       dell'opportuna azione
 *  </ul>
 * </li>
 *
 * <i>Nota</i>:
 *  necessario che questa classe implementi Runnable perch all'interno di una
 * midlet le operazioni bloccanti, quali quelle relative al networking, devono
 * essere eseguite in un  thread differente, al fine di evitare deadlock
 *
 * @author Antonio
 */
public abstract class RequestProcessor implements Runnable {
    
    private Thread processorThread;
    protected ServiceResponseHandler handler = null;
    protected DataOutputStream streamToServer = null;
    protected DataInputStream streamFromServer = null;
    
    
    
    
    /**
     * inoltra la richiesta alla midlet server
     */
    protected void sendRequest(String toSend) throws BtIOException{
        //TODO debug
        System.out.println("RequestProcessor::sendRequest toSend = " + toSend);
        
        try{
            OutputStream out = this.streamToServer;
            out.write(toSend.getBytes());
        } catch (IOException e){
            throw new BtIOException(e.getMessage());
        }
    }
    
    /**
     * riceve il response di una richiesta dalla midlet server
     */
    private String getResponse(String endOfResponseMarker) throws BtIOException{
        //TODO fare anche qui il giochetto dei marker
        String received = "";
        try{
            DataInputStream input = this.streamFromServer;
            byte buffer[] = new byte[endOfResponseMarker.length()];
            int byteReaded;
            String temp;
            boolean isEnded = false;
            
            do{
                byteReaded = input.read(buffer);
                temp = new String(buffer);
                System.out.println("RequestProcessor::sto ascoltando");
                
                //TODO debug
                System.out.println("REquest processor:: " + byteReaded);
                
                received += temp.substring(0, byteReaded);
                
                System.out.println("REquest processor:: received  " + received);
                
                /*
                 * quando  stata ricevuta la stringa che marca la fine del
                 * response (endOfResponseMarker) non devo fare pi alcuna read
                 */
                if (received.length()> endOfResponseMarker.length() ){
                    int startIndex = received.length()- endOfResponseMarker.length();
                    
                    //TODO se received contiene l'endMarker ma a dop questo ci sono dei whitespaces l'if di sotto non funziona, cio, non si rileva la ricezione dell'endmarker
                    if (received.substring(startIndex, received.length()).equalsIgnoreCase(endOfResponseMarker))
                        isEnded = true;
                }
            }
            
            while( !isEnded);
            
        } catch (IOException e){
            throw new BtIOException(e.getMessage());
        }
        
        //TODO debug
        System.out.println("REquest processor:: " + received);
        //TODO debug
        System.out.println("REquest processor:: stampa!" );
        
        return received;
        
    }
    
    protected String getResponseContent(String beginMarker, String endMarker) throws BtIOException{
        String out = getResponse(endMarker);
        
        out = out.substring(beginMarker.length(),out.length());
        int endIndex = out.length() - endMarker.length();
        out = out.substring(0, endIndex);
        return out;
    }
    
    
    /**
     * lancia in esecuzione il thread
     */
    public void  processRequest(ServiceResponseHandler handler, DataInputStream streamFromServer, DataOutputStream streamToServer){
        this.handler = handler;
        this.streamFromServer = streamFromServer;
        this.streamToServer = streamToServer;
        this.processorThread = new Thread(this);
        processorThread.start();
    }
    
}
